import time


class CostTimer:
    """Context manager to measure wall time"""

    def __enter__(self):
        self._start = time.perf_counter()
        return self

    def __exit__(self, *args):
        end = time.perf_counter()
        self.time = end - self._start



import cachey
import xarray as xr
from fastapi import FastAPI

from .rest import SingleDatasetRest


@xr.register_dataset_accessor('rest')
class RestAccessor:
    """REST API Accessor for serving one dataset in its
    dedicated FastAPI application.

    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._rest = None

        self._initialized = False

    def _get_rest_obj(self):
        if self._rest is None:
            self._rest = SingleDatasetRest(self._obj)

        return self._rest

    def __call__(self, **kwargs):
        """Initialize this accessor by setting optional configuration values.

        Parameters
        ----------
        **kwargs
            Arguments passed to :func:`xpublish.SingleDatasetRest.__init__`.

        Notes
        -----
        This method can only be invoked once.

        """
        if self._initialized:
            raise RuntimeError('This accessor has already been initialized')
        self._initialized = True

        self._rest = SingleDatasetRest(self._obj, **kwargs)

        return self

    @property
    def cache(self) -> cachey.Cache:
        """Returns the :class:`cachey.Cache` instance used by the FastAPI application."""

        return self._get_rest_obj().cache

    @property
    def app(self) -> FastAPI:
        """Returns the :class:`fastapi.FastAPI` application instance."""

        return self._get_rest_obj().app

    def serve(self, **kwargs):
        """Serve this FastAPI application via :func:`uvicorn.run`.

        Parameters
        ----------
        **kwargs :
            Arguments passed to :func:`xpublish.SingleDatasetRest.serve`.

        Notes
        -----
        This method is blocking and does not return.

        """
        self._get_rest_obj().serve(**kwargs)



import copy
import logging
from typing import (
    Any,
    Optional,
)

import dask.array
import numpy as np
import xarray as xr
from numcodecs.abc import Codec
from numcodecs.compat import ensure_ndarray
from xarray.backends.zarr import (
    DIMENSION_KEY,
    encode_zarr_attr_value,
    encode_zarr_variable,
    extract_zarr_variable_encoding,
)
from zarr.meta import encode_fill_value
from zarr.storage import (
    array_meta_key,
    attrs_key,
    default_compressor,
    group_meta_key,
)
from zarr.util import normalize_shape

from .api import DATASET_ID_ATTR_KEY

DaskArrayType = (dask.array.Array,)
ZARR_FORMAT = 2
ZARR_CONSOLIDATED_FORMAT = 1
ZARR_METADATA_KEY = '.zmetadata'

logger = logging.getLogger('api')


def _extract_dataset_zattrs(dataset: xr.Dataset) -> dict:
    """helper function to create zattrs dictionary from Dataset global attrs"""
    zattrs = {}
    for k, v in dataset.attrs.items():
        zattrs[k] = encode_zarr_attr_value(v)

    # remove xpublish internal attribute
    zattrs.pop(DATASET_ID_ATTR_KEY, None)

    return zattrs


def _extract_dataarray_zattrs(da: xr.DataArray) -> dict:
    """helper function to extract zattrs dictionary from DataArray"""
    zattrs = {}
    for k, v in da.attrs.items():
        zattrs[k] = encode_zarr_attr_value(v)
    zattrs[DIMENSION_KEY] = list(da.dims)

    # We don't want `_FillValue` in `.zattrs`
    # It should go in `fill_value` section of `.zarray`
    _ = zattrs.pop('_FillValue', None)

    return zattrs


def _extract_dataarray_coords(
    da: xr.DataArray,
    zattrs: dict,
) -> dict:
    '''helper function to extract coords from DataArray into a directionary'''
    if da.coords:
        # Coordinates are only encoded if there are non-dimension coordinates
        nondim_coords = set(da.coords) - set(da.dims)

        if len(nondim_coords) > 0 and da.name not in nondim_coords:
            coords = ' '.join(sorted(nondim_coords))
            zattrs['coordinates'] = encode_zarr_attr_value(coords)
    return zattrs


def _extract_fill_value(
    da: xr.DataArray,
    dtype: np.dtype,
) -> Any:
    """helper function to extract fill value from DataArray."""
    fill_value = da.attrs.pop('_FillValue', None)
    return encode_fill_value(fill_value, dtype)


def _extract_zarray(
    da: xr.DataArray,
    encoding: dict,
    dtype: np.dtype,
) -> dict:
    """helper function to extract zarr array metadata."""
    meta = {
        'compressor': encoding.get('compressor', da.encoding.get('compressor', default_compressor)),
        'filters': encoding.get('filters', da.encoding.get('filters', None)),
        'chunks': encoding.get('chunks', None),
        'dtype': dtype.str,
        'fill_value': _extract_fill_value(da, dtype),
        'order': 'C',
        'shape': list(normalize_shape(da.shape)),
        'zarr_format': ZARR_FORMAT,
    }

    if meta['chunks'] is None:
        meta['chunks'] = da.shape

    # validate chunks
    if isinstance(da.data, DaskArrayType):
        var_chunks = tuple([c[0] for c in da.data.chunks])
    else:
        var_chunks = da.shape
    if not var_chunks == tuple(meta['chunks']):
        raise ValueError('Encoding chunks do not match inferred chunks')

    meta['chunks'] = list(meta['chunks'])  # return chunks as a list

    return meta


def create_zvariables(dataset: xr.Dataset) -> dict:
    """Helper function to create a dictionary of zarr encoded variables."""
    zvariables = {}

    for key, da in dataset.variables.items():
        encoded_da = encode_zarr_variable(da, name=key)
        zvariables[key] = encoded_da

    return zvariables


def create_zmetadata(dataset: xr.Dataset) -> dict:
    """Helper function to create a consolidated zmetadata dictionary."""

    zmeta = {
        'zarr_consolidated_format': ZARR_CONSOLIDATED_FORMAT,
        'metadata': {},
    }
    zmeta['metadata'][group_meta_key] = {'zarr_format': ZARR_FORMAT}
    zmeta['metadata'][attrs_key] = _extract_dataset_zattrs(dataset)

    for key, dvar in dataset.variables.items():
        da = dataset[key]
        encoded_da = encode_zarr_variable(dvar, name=key)
        encoding = extract_zarr_variable_encoding(dvar)
        zattrs = _extract_dataarray_zattrs(encoded_da)
        zattrs = _extract_dataarray_coords(da, zattrs)
        zmeta['metadata'][f'{key}/{attrs_key}'] = zattrs
        zmeta['metadata'][f'{key}/{array_meta_key}'] = _extract_zarray(
            encoded_da,
            encoding,
            encoded_da.dtype,
        )

    return zmeta


def jsonify_zmetadata(
    dataset: xr.Dataset,
    zmetadata: dict,
) -> dict:
    """Helper function to convert zmetadata dictionary to a json
    compatible dictionary.

    """
    zjson = copy.deepcopy(zmetadata)

    for key in list(dataset.variables):
        # convert compressor to dict
        compressor = zjson['metadata'][f'{key}/{array_meta_key}']['compressor']
        if compressor is not None:
            compressor_config = zjson['metadata'][f'{key}/{array_meta_key}'][
                'compressor'
            ].get_config()
            zjson['metadata'][f'{key}/{array_meta_key}']['compressor'] = compressor_config

    return zjson


def encode_chunk(
    chunk: np.typing.ArrayLike,
    filters: Optional[list[Codec]] = None,
    compressor: Optional[Codec] = None,
) -> np.typing.ArrayLike:
    """helper function largely copied from zarr.Array"""
    # apply filters
    if filters:
        for f in filters:
            chunk = f.encode(chunk)

    # check object encoding
    if ensure_ndarray(chunk).dtype == object:
        raise RuntimeError('cannot write object array without object codec')

    # compress
    if compressor:
        cdata = compressor.encode(chunk)
    else:
        cdata = chunk

    return cdata


def get_data_chunk(
    da: xr.DataArray,
    chunk_id: str,
    out_shape: tuple,
) -> np.typing.ArrayLike:
    """Get one chunk of data from this DataArray (da).

    If this is an incomplete edge chunk, pad the returned array to match out_shape.
    """
    ikeys = tuple(map(int, chunk_id.split('.')))
    if isinstance(da, DaskArrayType):
        chunk_data = da.blocks[ikeys]
    else:
        if da.ndim > 0 and ikeys != ((0,) * da.ndim):
            raise ValueError(
                'Invalid chunk_id for numpy array: %s. Should have been: %s'
                % (chunk_id, ((0,) * da.ndim))
            )
        chunk_data = np.asarray(da)

    logger.debug('checking chunk output size, %s == %s' % (chunk_data.shape, out_shape))

    if isinstance(chunk_data, DaskArrayType):
        chunk_data = chunk_data.compute()

    # zarr expects full edge chunks, contents out of bounds for the array are undefined
    if chunk_data.shape != tuple(out_shape):
        new_chunk = np.empty_like(chunk_data, shape=out_shape)
        write_slice = tuple([slice(0, s) for s in chunk_data.shape])
        new_chunk[write_slice] = chunk_data
        return new_chunk
    else:
        return chunk_data



"""
Version information router
"""
import importlib
import sys
from typing import Sequence

from fastapi import APIRouter

from ...utils.info import get_sys_info, netcdf_and_hdf5_versions
from .. import Plugin, hookimpl


class ModuleVersionPlugin(Plugin):
    """Share the currently loaded versions of key libraries"""

    name: str = 'module_version'

    app_router_prefix: str = ''
    app_router_tags: Sequence[str] = ['module_version']

    @hookimpl
    def app_router(self) -> APIRouter:
        router = APIRouter(
            prefix=self.app_router_prefix,
            tags=self.app_router_tags,
        )

        @router.get('/versions')
        def get_versions() -> dict:
            """Currently loaded versions of key libraries"""
            versions = dict(get_sys_info() + netcdf_and_hdf5_versions())
            modules = [
                'xarray',
                'zarr',
                'numcodecs',
                'fastapi',
                'starlette',
                'pandas',
                'numpy',
                'dask',
                'distributed',
                'uvicorn',
            ]
            for modname in modules:
                try:
                    if modname in sys.modules:
                        mod = sys.modules[modname]
                    else:  # pragma: no cover
                        mod = importlib.import_module(modname)
                    versions[modname] = getattr(mod, '__version__', None)
                except ImportError:  # pragma: no cover
                    pass
            return versions

        return router



import json
from collections.abc import Mapping
from typing import Any, Dict, List, Tuple

import xarray as xr
from fastapi import APIRouter
from fastapi.openapi.utils import get_openapi
from starlette.responses import JSONResponse as StarletteJSONResponse  # type: ignore

DATASET_ID_ATTR_KEY = '_xpublish_id'


def normalize_datasets(datasets) -> Dict[str, xr.Dataset]:
    """Normalize the given collection of datasets.

    - raise TypeError if objects other than xarray.Dataset are found
    - return an empty dictionary in the special case where a single dataset is given
    - convert all keys (dataset ids) to strings
    - add dataset ids to their corresponding dataset object as global attribute
      (so that it can be easily retrieved within path operation functions).

    """
    error_msg = 'Can only publish a xarray.Dataset object or a mapping of Dataset objects'

    if isinstance(datasets, xr.Dataset):
        return {}
    elif isinstance(datasets, Mapping):
        if not all(isinstance(obj, xr.Dataset) for obj in datasets.values()):
            raise TypeError(error_msg)
        return {str(k): ds.assign_attrs({DATASET_ID_ATTR_KEY: k}) for k, ds in datasets.items()}
    else:
        raise TypeError(error_msg)


def normalize_app_routers(
    routers: list,
    prefix: str,
) -> List[Tuple[APIRouter, Dict]]:
    """Normalise the given list of (dataset-specific) API routers.

    Add or prepend ``prefix`` to all routers.

    """
    new_routers = []

    for rt in routers:
        if isinstance(rt, APIRouter):
            new_routers.append((rt, {'prefix': prefix}))
        elif isinstance(rt, tuple) and isinstance(rt[0], APIRouter) and len(rt) == 2:
            rt_kwargs = rt[1]
            rt_kwargs['prefix'] = prefix + rt_kwargs.get('prefix', '')
            new_routers.append((rt[0], rt_kwargs))
        else:
            raise TypeError(
                'Invalid type/format for routers argument, please provide either an APIRouter '
                'instance or a (APIRouter, {...}) tuple.'
            )

    return new_routers


def check_route_conflicts(routers) -> None:
    paths = []

    for router, kws in routers:
        prefix = kws.get('prefix', '')
        paths += [prefix + r.path for r in router.routes]

    seen = set()
    duplicates = []

    for p in paths:
        if p in seen:
            duplicates.append(p)
        else:
            seen.add(p)

    if len(duplicates):
        raise ValueError(f'Found multiple routes defined for the following paths: {duplicates}')


class SingleDatasetOpenAPIOverrider:
    """Used to override the FastAPI application openapi specs when a single
    dataset is published.

    In this case, the "dataset_id" path parameter is not present in API
    endpoints and has to be removed manually.

    See:

    - https://fastapi.tiangolo.com/advanced/extending-openapi/
    - https://github.com/tiangolo/fastapi/issues/1594

    """

    def __init__(self, app) -> None:
        self._app = app

    def openapi(self) -> dict:
        if self._app.openapi_schema:
            return self._app.openapi_schema

        kwargs = {
            'title': self._app.title,
            'version': self._app.version,
            'description': self._app.description,
            'routes': self._app.routes,
            'tags': self._app.openapi_tags,
            'servers': self._app.servers,
        }

        openapi_schema = get_openapi(**kwargs)

        for path in openapi_schema.get('paths', {}).values():
            for http_method in path.values():
                params = http_method.get('parameters')

                if params is not None:
                    for i, p in enumerate(params):
                        if p.get('name') == 'dataset_id':
                            params.pop(i)

        self._app.openapi_schema = openapi_schema

        return self._app.openapi_schema


class JSONResponse(StarletteJSONResponse):
    def __init__(self, *args, **kwargs) -> None:
        self._render_kwargs = {
            'ensure_ascii': True,
            'allow_nan': True,
            'indent': None,
            'separators': (',', ':'),
        }
        self._render_kwargs.update(kwargs.pop('render_kwargs', {}))
        super().__init__(*args, **kwargs)

    def render(self, content: Any) -> bytes:
        return json.dumps(content, **self._render_kwargs).encode('utf-8')



"""
Plugin information router
"""
import importlib
from typing import Dict, Optional, Sequence

from fastapi import APIRouter, Depends
from pydantic import BaseModel

from .. import Dependencies, Plugin, hookimpl


class PluginInfo(BaseModel):
    path: str
    version: Optional[str] = None


class PluginInfoPlugin(Plugin):
    """Expose plugin source and version"""

    name: str = 'plugin_info'

    app_router_prefix: str = ''
    app_router_tags: Sequence[str] = ['plugin_info']

    @hookimpl
    def app_router(self, deps: Dependencies) -> APIRouter:
        router = APIRouter(
            prefix=self.app_router_prefix,
            tags=list(self.app_router_tags),
        )

        @router.get('/plugins')
        def get_plugins(
            plugins: Dict[str, Plugin] = Depends(deps.plugins)
        ) -> Dict[str, PluginInfo]:
            """Return the source and version of the currently loaded plugins"""
            plugin_info = {}

            for name, plugin in plugins.items():
                plugin_type = type(plugin)
                module_name = plugin_type.__module__.split('.')[0]
                try:
                    mod = importlib.import_module(module_name)
                    version = getattr(mod, '__version__', None)
                except ImportError:  # pragma: no cover
                    version = None  # pragma: no cover

                plugin_info[name] = PluginInfo(
                    path=f'{plugin_type.__module__}.{plugin.__repr_name__()}',
                    version=version,
                )

            return plugin_info

        return router



from typing import Dict, List, Optional, Tuple

import cachey
import pluggy
import uvicorn
import xarray as xr
from fastapi import APIRouter, FastAPI, HTTPException, Path

from .dependencies import get_cache, get_dataset, get_dataset_ids, get_plugin_manager
from .plugins import Dependencies, Plugin, PluginSpec, get_plugins, load_default_plugins
from .routers import dataset_collection_router
from .utils.api import (
    SingleDatasetOpenAPIOverrider,
    check_route_conflicts,
    normalize_app_routers,
    normalize_datasets,
)

RouterKwargs = Dict
RouterAndKwargs = Tuple[APIRouter, RouterKwargs]


class Rest:
    """Used to publish multiple Xarray Datasets via a REST API (FastAPI application).

    To publish a single dataset via its own FastAPI application, you might
    want to use the :attr:`xarray.Dataset.rest` accessor for more convenience.
    Additionally the :class:`xpublish.SingleDatasetRest` class allows has
    a simplified interface for single dataset access.

    Parameters
    ----------
    datasets :
        A mapping of datasets objects to be served. If a mapping is given, keys
        are used as dataset ids and are converted to strings. See also the notes below.
    routers :
        A list of dataset-specific :class:`fastapi.APIRouter` instances to
        include in the fastAPI application. These routers are in addition
        to any loaded via plugins.
        The items of the list may also be tuples with the following format:
        ``[(router1, {'prefix': '/foo', 'tags': ['foo', 'bar']})]``, where
        the 1st tuple element is a :class:`fastapi.APIRouter` instance and the
        2nd element is a dictionary that is used to pass keyword arguments to
        :meth:`fastapi.FastAPI.include_router`.
    cache_kws :
        Dictionary of keyword arguments to be passed to
        :meth:`cachey.Cache.__init__()`.
        By default, the cache size is set to 1MB, but this can be changed with
        ``available_bytes``.
    app_kws :
        Dictionary of keyword arguments to be passed to
        :meth:`fastapi.FastAPI.__init__()`.
    plugins :
        Optional dictionary of loaded, configured plugins.
        Overrides automatic loading of plugins.
        If no plugins are desired, set to an empty dict.

    Notes
    -----
    The urls of the application's API endpoints differ whether a single dataset
    or a mapping (collection) of datasets is given. In the latter case, all
    dataset-specific endpoint urls have the prefix ``/datasets/{dataset_id}``,
    where ``{dataset_id}`` corresponds to the keys of the mapping (converted to
    strings). Still in the latter case, the endpoint ``/datasets`` is added and returns
    the list of all dataset ids.

    """

    def __init__(
        self,
        datasets: Optional[Dict[str, xr.Dataset]] = None,
        routers: Optional[APIRouter] = None,
        cache_kws: Optional[Dict] = None,
        app_kws: Optional[Dict] = None,
        plugins: Optional[Dict[str, Plugin]] = None,
    ):
        if isinstance(datasets, xr.Dataset):
            raise TypeError(
                'xpublish.Rest no longer directly handles single datasets. Please use xpublish.SingleDatasetRest instead'
            )

        self.setup_datasets(datasets or {})
        self.setup_plugins(plugins)

        routers = normalize_app_routers(
            routers or [],
            self._dataset_route_prefix,
        )
        check_route_conflicts(routers)
        self._routers = routers

        self.init_app_kwargs(app_kws)
        self.init_cache_kwargs(cache_kws)

    def setup_datasets(self, datasets: Dict[str, xr.Dataset]) -> str:
        """Initialize datasets and dataset accessor function

        Returns:
            Prefix for dataset routers
        """
        self._datasets = normalize_datasets(datasets)

        self._get_dataset_func = self.get_dataset_from_plugins
        self._dataset_route_prefix = '/datasets/{dataset_id}'
        return self._dataset_route_prefix

    def get_datasets_from_plugins(self) -> List[str]:
        """Return dataset ids from directly loaded datasets and plugins

        Used as a FastAPI dependency in dataset router plugins
        via :meth:`Rest.dependencies`.

        Returns:
            Dataset IDs from plugins and datasets loaded into
            :class:`xpublish.Rest` at initialization.
        """
        dataset_ids = list(self._datasets)

        for plugin_dataset_ids in self.pm.hook.get_datasets():
            dataset_ids.extend(plugin_dataset_ids)

        return dataset_ids

    def get_dataset_from_plugins(
        self,
        dataset_id: str = Path(description='Unique ID of dataset'),
    ) -> xr.Dataset:
        """Attempt to load dataset from plugins, otherwise return dataset from passed in dictionary of datasets

        Parameters:
            dataset_id:
                Unique key of dataset to attempt to load from plugins or
                those provided to :class:`xpublish.Rest` at initialization.

        Returns:
            Dataset for selected ``dataset_id``

        Raises:
            FastAPI.HTTPException: When a dataset is not found a 404 error is returned.
        """
        dataset = self.pm.hook.get_dataset(dataset_id=dataset_id)

        if dataset:
            return dataset

        if dataset_id not in self._datasets:
            raise HTTPException(status_code=404, detail=f"Dataset '{dataset_id}' not found")

        return self._datasets[dataset_id]

    def setup_plugins(
        self,
        plugins: Optional[Dict[str, Plugin]] = None,
    ) -> None:
        """Initialize and load plugins from entry_points unless explicitly provided

        Parameters:
            plugins:
                If a dictionary of initialized plugins is provided,
                then the automatic loading of plugins is disabled.

                Providing an empty dictionary will also disable
                automatic loading of plugins.
        """
        if plugins is None:
            plugins = load_default_plugins()

        self.pm = pluggy.PluginManager('xpublish')
        self.pm.add_hookspecs(PluginSpec)

        for name, plugin in plugins.items():
            self.pm.register(plugin, name=name)

        for hookspec in self.pm.hook.register_hookspec():
            self.pm.add_hookspecs(hookspec)

    def register_plugin(
        self,
        plugin: Plugin,
        plugin_name: Optional[str] = None,
        overwrite: bool = False,
    ) -> None:
        """
        Register a plugin with the xpublish system

        Args:
            plugin: Instantiated Plugin object
            plugin_name: Plugin name
            overwrite: If a plugin of the same name exist,
                setting this to True will remove the existing plugin before
                registering the new plugin. Defaults to False.

        Raises:
            AttributeError: Plugin can not be registered
            ValueError: Plugin already registered, try setting overwrite to True
        """
        plugin_name = plugin_name or plugin.name

        if overwrite is True and plugin_name in dict(self.pm.list_name_plugin()):
            # If a plugin exist with the same name, unregister it.
            # If configured using entry_points, the name of the
            # entry_point should be the same as the plugin.name.
            self.pm.unregister(name=plugin_name)

        # Get existing plugins again
        existing_plugins = self.pm.get_plugins()
        try:
            self.pm.register(plugin, plugin_name)
        except AttributeError as e:
            raise AttributeError(
                f'Plugin {plugin} is likely not initialized before registration'
            ) from e

        for hookspec in self.pm.subset_hook_caller(
            'register_hookspec', remove_plugins=existing_plugins
        )():
            self.pm.add_hookspecs(hookspec)

    def init_cache_kwargs(self, cache_kws: dict) -> None:
        """Set up cache kwargs"""
        self._cache = None
        self._cache_kws = {'available_bytes': 1e6}
        if cache_kws is not None:
            self._cache_kws.update(cache_kws)

    def init_app_kwargs(self, app_kws: dict) -> None:
        """Set up FastAPI application kwargs"""
        self._app = None
        self._app_kws = {}
        if app_kws is not None:
            self._app_kws.update(app_kws)

    @property
    def cache(self) -> cachey.Cache:
        """Returns the :class:`cachey.Cache` instance used by the FastAPI application."""

        if self._cache is None:
            self._cache = cachey.Cache(**self._cache_kws)
        return self._cache

    @property
    def plugins(self) -> Dict[str, Plugin]:
        """Returns the loaded plugins"""
        return dict(self.pm.list_name_plugin())

    def _init_routers(self, dataset_routers: Optional[APIRouter]) -> None:
        """Setup plugin and dataset routers. Needs to run after dataset and plugin setup"""
        app_routers, plugin_dataset_routers = self.plugin_routers()

        if self._dataset_route_prefix:
            app_routers.append((dataset_collection_router, {'tags': ['info']}))

        app_routers.extend(
            normalize_app_routers(
                plugin_dataset_routers + (dataset_routers or []), self._dataset_route_prefix
            )
        )

        check_route_conflicts(app_routers)

        self._app_routers = app_routers

    def plugin_routers(self) -> Tuple[List[RouterAndKwargs], List[RouterAndKwargs]]:
        """Load the app and dataset routers for plugins

        Returns:
            A tuple containing a list of top-level routers from plugins
            and a list of per-dataset routers from plugins
        """
        app_routers = []
        dataset_routers = []

        deps = self.dependencies()

        for router in self.pm.hook.app_router(deps=deps):
            app_routers.append((router, {}))

        for router in self.pm.hook.dataset_router(deps=deps):
            dataset_routers.append((router, {}))

        return app_routers, dataset_routers

    def dependencies(self) -> Dependencies:
        """FastAPI dependencies to pass to plugin router methods"""
        deps = Dependencies(
            dataset_ids=self.get_datasets_from_plugins,
            dataset=self._get_dataset_func,
            cache=lambda: self.cache,
            plugins=lambda: self.plugins,
            plugin_manager=lambda: self.pm,
        )

        return deps

    def _init_dependencies(self) -> None:
        """Initialize dependencies"""
        deps = self.dependencies()

        self._app.dependency_overrides[get_dataset_ids] = deps.dataset_ids
        self._app.dependency_overrides[get_dataset] = deps.dataset
        self._app.dependency_overrides[get_cache] = deps.cache
        self._app.dependency_overrides[get_plugins] = deps.plugins
        self._app.dependency_overrides[get_plugin_manager] = deps.plugin_manager

    def _init_app(self) -> FastAPI:
        """Initiate the FastAPI application."""

        self._app = FastAPI(**self._app_kws)

        self._init_routers(self._routers)
        for rt, kwargs in self._app_routers:
            self._app.include_router(rt, **kwargs)

        self._init_dependencies()

        return self._app

    @property
    def app(self) -> FastAPI:
        """Returns the :class:`fastapi.FastAPI` application instance.

        Notes
        -----
        Plugins registered with :meth:`xpublish.Rest.register_plugin` after :meth:`xpublish.Rest.app`
        is accessed or :meth:`xpublish.Rest.serve` is called once may not take effect.
        """
        if self._app is None:
            self._app = self._init_app()
        return self._app

    def serve(
        self,
        host: str = '0.0.0.0',
        port: int = 9000,
        log_level: str = 'debug',
        **kwargs,
    ) -> None:
        """Serve this FastAPI application via :func:`uvicorn.run`.

        Parameters
        ----------
        host :
            Bind socket to this host.
        port :
            Bind socket to this port.
        log_level :
            App logging level, valid options are
            {'critical', 'error', 'warning', 'info', 'debug', 'trace'}.
        **kwargs :
            Additional arguments to be passed to :func:`uvicorn.run`.

        Notes
        -----
        This method is blocking and does not return.

        """
        uvicorn.run(
            self.app,
            host=host,
            port=port,
            log_level=log_level,
            **kwargs,
        )


class SingleDatasetRest(Rest):
    """Used to publish a single Xarray dataset via a REST API (FastAPI application).
    Use :class:`xpublish.Rest` to publish multiple datasets.

    Parameters:
    -----------
    dataset :
        A single :class:`xarray.Dataset` object to be served.
    """

    def __init__(
        self,
        dataset: xr.Dataset,
        routers: Optional[APIRouter] = None,
        cache_kws: Optional[Dict] = None,
        app_kws: Optional[Dict] = None,
        plugins: Optional[Dict[str, Plugin]] = None,
    ):
        self._dataset = dataset

        super().__init__({}, routers, cache_kws, app_kws, plugins)

    def setup_datasets(self, datasets) -> str:
        """Modifies the dataset loading to instead connect to the
        single dataset"""
        self._dataset_route_prefix = ''
        self._datasets = {}

        self._get_dataset_func = lambda: self._dataset

        return self._dataset_route_prefix

    def _init_app(self) -> FastAPI:
        self._app = super()._init_app()

        self._app.openapi = SingleDatasetOpenAPIOverrider(self._app).openapi

        return self._app



"""Utility functions for printing version information, adapted from
xarray/util/print_versions.py

"""
import locale
import os
import platform
import struct
import subprocess
import sys
from typing import Union


def get_sys_info() -> list:
    'Returns system information as a dict'

    blob = []

    # get full commit hash
    if os.path.isdir('.git') and os.path.isdir('xpublish'):
        try:
            pipe = subprocess.Popen(
                'git log --format="%H" -n 1'.split(' '),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            so, serr = pipe.communicate()
        except Exception:  # pragma: no cover
            commit = None
        else:
            if pipe.returncode == 0:
                commit = so
                try:
                    commit = so.decode('utf-8')
                except ValueError:  # pragma: no cover
                    pass
                commit = commit.strip().strip('"')

        blob.append(('commit', commit))

    uname = platform.uname()
    blob.extend(
        [
            ('python', sys.version),
            ('python-bits', struct.calcsize('P') * 8),
            ('OS', '%s' % (uname.system)),
            ('OS-release', '%s' % (uname.release)),
            ('Version', '%s' % (uname.version)),
            ('machine', '%s' % (uname.machine)),
            ('processor', '%s' % (uname.processor)),
            ('byteorder', '%s' % sys.byteorder),
            ('LC_ALL', '%s' % os.environ.get('LC_ALL', 'None')),
            ('LANG', '%s' % os.environ.get('LANG', 'None')),
            ('LOCALE', '%s.%s' % locale.getlocale()),
        ]
    )

    return blob


def netcdf_and_hdf5_versions() -> list[tuple[str, Union[str, None]]]:
    libhdf5_version = None
    libnetcdf_version = None
    try:
        import netCDF4

        libhdf5_version = netCDF4.__hdf5libversion__
        libnetcdf_version = netCDF4.__netcdf4libversion__
    except ImportError:  # pragma: no cover
        try:
            import h5py

            libhdf5_version = h5py.version.hdf5_version
        except ImportError:
            pass
    return [('libhdf5', libhdf5_version), ('libnetcdf', libnetcdf_version)]



from typing import Callable, Dict, Iterable, List, Optional

import cachey  # type: ignore
import pluggy  # type: ignore
import xarray as xr
from fastapi import APIRouter
from pydantic import BaseModel, Field

from ..dependencies import get_cache, get_dataset, get_dataset_ids, get_plugin_manager, get_plugins

# Decorator helper to mark functions as Xpublish hook specifications
hookspec = pluggy.HookspecMarker('xpublish')

# Decorator helper to mark functions as Xpublish hook implementations
hookimpl = pluggy.HookimplMarker('xpublish')


class Dependencies(BaseModel):
    """
    A set of dependencies that are passed into plugin routers.

    Some routers may be 'borrowed' by other routers to expose different
    geometries of data, thus the default dependencies may need to be overridden.
    By depending on the passed in version of this class, the dependencies
    can be overridden predictably.
    """

    dataset_ids: Callable[..., List[str]] = Field(
        get_dataset_ids,
        description='Returns a list of all valid dataset ids',
    )
    dataset: Callable[[str], xr.Dataset] = Field(
        get_dataset,
        description='Returns a dataset using ``/<dataset_id>/`` in the path.',
    )
    cache: Callable[..., cachey.Cache] = Field(
        get_cache,
        description='Provide access to :py:class:`cachey.Cache`',
    )
    plugins: Callable[..., Dict[str, 'Plugin']] = Field(
        get_plugins,
        description='A dictionary of plugins allowing direct access',
    )
    plugin_manager: Callable[..., pluggy.PluginManager] = Field(
        get_plugin_manager,
        description='The plugin manager itself, allowing for maximum creativity',
    )

    def __hash__(self):
        """Dependency functions aren't easy to hash"""
        return 0  # pragma: no cover


class Plugin(BaseModel):
    """
    Xpublish plugins provide ways to extend the core of xpublish with
    new routers and other functionality.

    To create a plugin, subclass `Plugin` and add attributes that are
    subclasses of `PluginType` (`Router` for instance).

    The specific attributes correspond to how Xpublish should use
    the plugin.
    """

    name: str = Field(..., description='Fallback name of plugin')

    def __hash__(self):
        """Make sure that the plugin is hashable to load with pluggy"""
        things_to_hash = []

        # try/except is for pydantic backwards compatibility
        try:
            model_dict = self.model_dump()
        except AttributeError:
            model_dict = self.dict()

        for e in model_dict:
            if isinstance(e, list):
                things_to_hash.append(tuple(e))  # pragma: no cover
            else:
                things_to_hash.append(e)

        return hash(tuple(things_to_hash))

    def __dir__(self) -> Iterable[str]:
        """We need to override the dir as pluggy will otherwise try to inspect it,
        and Pydantic has marked it class only

        https://github.com/pydantic/pydantic/pull/1466
        """
        d = list(super().__dir__())

        d.remove('__signature__')

        return d


class PluginSpec(Plugin):
    """Plugin extension points

    Plugins do not need to implement all of the methods defined here,
    instead they implement
    """

    @hookspec
    def app_router(self, deps: Dependencies) -> APIRouter:  # type: ignore
        """Create an app (top-level) router for the plugin

        Implementations should return an APIRouter, and define
        app_router_prefix, and app_router_tags on the class,
        and use those to initialize the router.
        """

    @hookspec
    def dataset_router(self, deps: Dependencies) -> APIRouter:  # type: ignore
        """Create a dataset router for the plugin

        Implementations should return an APIRouter, and define
        dataset_router_prefix, and dataset_router_tags on the class,
        and use those to initialize the router.
        """

    @hookspec
    def get_datasets(self) -> Iterable[str]:  # type: ignore
        """Return an iterable of dataset ids that the plugin can provide"""

    @hookspec(firstresult=True)
    # type: ignore
    def get_dataset(self, dataset_id: str) -> Optional[xr.Dataset]:
        """Return a dataset by requested dataset_id.

        If the plugin does not have the dataset, return None
        """

    @hookspec
    def register_hookspec(self):  # type: ignore
        """Return additional hookspec class to register with the plugin manager"""



"""
Dataset-independent API routes.

"""
from fastapi import APIRouter, Depends

from ..dependencies import get_dataset_ids

dataset_collection_router = APIRouter()


@dataset_collection_router.get('/datasets')
def get_dataset_collection_keys(ids: list = Depends(get_dataset_ids)) -> list[str]:
    """Return all the currently known Dataset IDs"""
    return ids



import logging
from typing import Sequence

import cachey  # type: ignore
import xarray as xr
from fastapi import APIRouter, Depends, HTTPException, Path
from starlette.responses import Response  # type: ignore
from zarr.storage import array_meta_key, attrs_key, group_meta_key  # type: ignore

from xpublish.utils.api import JSONResponse

from ...dependencies import get_zmetadata, get_zvariables
from ...utils.api import DATASET_ID_ATTR_KEY
from ...utils.cache import CostTimer
from ...utils.zarr import (
    ZARR_METADATA_KEY,
    encode_chunk,
    get_data_chunk,
    jsonify_zmetadata,
)
from .. import Dependencies, Plugin, hookimpl

logger = logging.getLogger('zarr_api')


class ZarrPlugin(Plugin):
    """Adds Zarr-like accessing endpoints for datasets"""

    name: str = 'zarr'

    dataset_router_prefix: str = '/zarr'
    dataset_router_tags: Sequence[str] = ['zarr']

    @hookimpl
    def dataset_router(self, deps: Dependencies) -> APIRouter:
        router = APIRouter(
            prefix=self.dataset_router_prefix,
            tags=list(self.dataset_router_tags),
        )

        @router.get(f'/{ZARR_METADATA_KEY}')
        def get_zarr_metadata(
            dataset=Depends(deps.dataset),
            cache=Depends(deps.cache),
        ) -> dict:
            """Consolidated Zarr metadata"""
            zvariables = get_zvariables(dataset, cache)
            zmetadata = get_zmetadata(dataset, cache, zvariables)

            zjson = jsonify_zmetadata(dataset, zmetadata)

            return JSONResponse(zjson)

        @router.get(f'/{group_meta_key}')
        def get_zarr_group(
            dataset=Depends(deps.dataset),
            cache=Depends(deps.cache),
        ) -> dict:
            """Zarr group data"""
            zvariables = get_zvariables(dataset, cache)
            zmetadata = get_zmetadata(dataset, cache, zvariables)

            return JSONResponse(zmetadata['metadata'][group_meta_key])

        @router.get(f'/{attrs_key}')
        def get_zarr_attrs(
            dataset=Depends(deps.dataset),
            cache=Depends(deps.cache),
        ) -> dict:
            """Zarr attributes"""
            zvariables = get_zvariables(dataset, cache)
            zmetadata = get_zmetadata(dataset, cache, zvariables)

            return JSONResponse(zmetadata['metadata'][attrs_key])

        @router.get('/{var}/{chunk}')
        def get_variable_chunk(
            var: str = Path(description='Variable in dataset'),
            chunk: str = Path(description='Zarr chunk'),
            dataset: xr.Dataset = Depends(deps.dataset),
            cache: cachey.Cache = Depends(deps.cache),
        ):
            """Get a zarr array chunk.

            This will return cached responses when available.

            """
            zvariables = get_zvariables(dataset, cache)
            zmetadata = get_zmetadata(dataset, cache, zvariables)

            # First check that this request wasn't for variable metadata
            if array_meta_key in chunk:
                return zmetadata['metadata'][f'{var}/{array_meta_key}']
            elif attrs_key in chunk:
                return JSONResponse(zmetadata['metadata'][f'{var}/{attrs_key}'])
            elif group_meta_key in chunk:
                raise HTTPException(status_code=404, detail='No subgroups')
            else:
                logger.debug('var is %s', var)
                logger.debug('chunk is %s', chunk)

                cache_key = dataset.attrs.get(DATASET_ID_ATTR_KEY, '') + '/' + f'{var}/{chunk}'
                response = cache.get(cache_key)

                if response is None:
                    with CostTimer() as ct:
                        arr_meta = zmetadata['metadata'][f'{var}/{array_meta_key}']
                        da = zvariables[var].data

                        data_chunk = get_data_chunk(
                            da,
                            chunk,
                            out_shape=arr_meta['chunks'],
                        )

                        echunk = encode_chunk(
                            data_chunk.tobytes(),
                            filters=arr_meta['filters'],
                            compressor=arr_meta['compressor'],
                        )

                        response = Response(
                            echunk,
                            media_type='application/octet-stream',
                        )

                    cache.put(cache_key, response, ct.time, len(echunk))

                return response

        return router



"""
Load and configure Xpublish plugins from entry point group `xpublish.plugin`
"""
from importlib.metadata import entry_points
from typing import Dict, Iterable, Optional, Type

from .hooks import Plugin


def find_default_plugins(
    exclude_plugins: Optional[Iterable[str]] = None,
) -> Dict[str, Type[Plugin]]:
    """Find Xpublish plugins from entry point group `xpublish.plugin`

    Individual plugins may be ignored by adding them to `exclude_plugins`.
    """
    exclude_plugins = set(exclude_plugins or [])

    plugins: Dict[str, Type[Plugin]] = {}

    for entry_point in entry_points()['xpublish.plugin']:
        if entry_point.name not in exclude_plugins:
            plugins[entry_point.name] = entry_point.load()

    return plugins


def load_default_plugins(
    exclude_plugins: Optional[Iterable[str]] = None,
) -> Dict[str, Plugin]:
    """Find and initialize plugins from entry point group `xpublish.plugin`"""
    initialized_plugins: Dict[str, Plugin] = {}

    for name, plugin in find_default_plugins(exclude_plugins=exclude_plugins).items():
        initialized_plugins[name] = plugin()

    return initialized_plugins


def configure_plugins(
    plugins: Dict[str, Type[Plugin]],
    plugin_configs: Optional[Dict] = None,
) -> Dict[str, Plugin]:
    """Initialize and configure plugins with given dictionary of configurations"""
    initialized_plugins: Dict[str, Plugin] = {}
    plugin_configs = plugin_configs or {}

    for name, plugin in plugins.items():
        kwargs = plugin_configs.get(name, {})
        initialized_plugins[name] = plugin(**kwargs)

    return initialized_plugins



from pkg_resources import DistributionNotFound, get_distribution

from .accessor import RestAccessor  # noqa: F401
from .plugins import Dependencies, Plugin, hookimpl, hookspec  # noqa: F401
from .rest import Rest, SingleDatasetRest  # noqa: F401

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # noqa: F401; pragma: no cover
    # package is not installed
    pass



"""
Helper functions to use a FastAPI dependencies.
"""
from typing import TYPE_CHECKING, Dict, List

import cachey
import pluggy
import xarray as xr
from fastapi import Depends

from .utils.api import DATASET_ID_ATTR_KEY
from .utils.zarr import ZARR_METADATA_KEY, create_zmetadata, create_zvariables

if TYPE_CHECKING:
    from .plugins import Plugin  # pragma: no cover


def get_dataset_ids() -> List[str]:
    """FastAPI dependency for getting the list of ids (string keys)
    of the collection of datasets being served.

    Use this callable as dependency in any FastAPI path operation
    function where you need access to those ids.

    This dummy dependency will be overridden when creating the FastAPI
    application.

    Returns:
        A list of unique keys for datasets

    """
    return []  # pragma: no cover


def get_dataset(dataset_id: str) -> xr.Dataset:
    """FastAPI dependency for accessing the published xarray dataset object.

    Use this callable as dependency in any FastAPI path operation
    function where you need access to the xarray Dataset being served.

    This dummy dependency will be overridden when creating the FastAPI
    application.

    Parameters:
        dataset_id:
            Unique path-safe key identifying dataset

    Returns:
        Requested Xarray dataset

    """
    return xr.Dataset()  # pragma: no cover


def get_cache() -> cachey.Cache:
    """FastAPI dependency for accessing the application's cache.

    Use this callable as dependency in any FastAPI path operation
    function where you need access to the cache provided with the
    application.

    This dummy dependency will be overridden when creating the FastAPI
    application.

    """
    return cachey.Cache(available_bytes=1e6)  # pragma: no cover


def get_zvariables(
    dataset: xr.Dataset = Depends(get_dataset), cache: cachey.Cache = Depends(get_cache)
):
    """FastAPI dependency that returns a dictionary of zarr encoded variables."""

    cache_key = dataset.attrs.get(DATASET_ID_ATTR_KEY, '') + '/' + 'zvariables'
    zvariables = cache.get(cache_key)

    if zvariables is None:
        zvariables = create_zvariables(dataset)

        # we want to permanently cache this: set high cost value
        cache.put(cache_key, zvariables, 99999)

    return zvariables


def get_zmetadata(
    dataset: xr.Dataset = Depends(get_dataset),
    cache: cachey.Cache = Depends(get_cache),
    zvariables: dict = Depends(get_zvariables),
):
    """FastAPI dependency that returns a consolidated zmetadata dictionary."""

    cache_key = dataset.attrs.get(DATASET_ID_ATTR_KEY, '') + '/' + ZARR_METADATA_KEY
    zmeta = cache.get(cache_key)

    if zmeta is None:
        zmeta = create_zmetadata(dataset)

        # we want to permanently cache this: set high cost value
        cache.put(cache_key, zmeta, 99999)

    return zmeta


def get_plugins() -> Dict[str, 'Plugin']:
    """FastAPI dependency that returns the a dictionary of loaded plugins

    Returns:
        Dictionary of names to initialized plugins.
    """

    return {}  # pragma: no cover


def get_plugin_manager() -> pluggy.PluginManager:
    """Return the active plugin manager"""



from typing import Sequence

import xarray as xr
from fastapi import APIRouter, Depends
from starlette.responses import HTMLResponse  # type: ignore
from zarr.storage import attrs_key  # type: ignore

from xpublish.utils.api import JSONResponse

from ...dependencies import get_zmetadata, get_zvariables
from .. import Dependencies, Plugin, hookimpl


class DatasetInfoPlugin(Plugin):
    """Dataset metadata"""

    name: str = 'dataset_info'

    dataset_router_prefix: str = ''
    dataset_router_tags: Sequence[str] = ['dataset_info']

    @hookimpl
    def dataset_router(self, deps: Dependencies) -> APIRouter:
        router = APIRouter(
            prefix=self.dataset_router_prefix,
            tags=list(self.dataset_router_tags),
        )

        @router.get('/')
        def html_representation(
            dataset=Depends(deps.dataset),
        ) -> HTMLResponse:
            """Returns the xarray HTML representation of the dataset."""

            with xr.set_options(display_style='html'):
                return HTMLResponse(dataset._repr_html_())

        @router.get('/keys')
        def list_keys(
            dataset=Depends(deps.dataset),
        ) -> list[str]:
            """List of the keys in a dataset"""

            return JSONResponse(list(dataset.variables))

        @router.get('/dict')
        def to_dict(
            dataset=Depends(deps.dataset),
        ) -> dict:
            """The full dataset as a dictionary"""
            return JSONResponse(dataset.to_dict(data=False))

        @router.get('/info')
        def info(
            dataset=Depends(deps.dataset),
            cache=Depends(deps.cache),
        ) -> dict:
            """Dataset schema (close to the NCO-JSON schema)."""

            zvariables = get_zvariables(dataset, cache)
            zmetadata = get_zmetadata(dataset, cache, zvariables)

            info = {}
            info['dimensions'] = dict(dataset.dims.items())
            info['variables'] = {}

            meta = zmetadata['metadata']

            for name, var in zvariables.items():
                attrs = meta[f'{name}/{attrs_key}'].copy()
                attrs.pop('_ARRAY_DIMENSIONS')

                info['variables'][name] = {
                    'type': var.data.dtype.name,
                    'dimensions': list(var.dims),
                    'attributes': attrs,
                }

            info['global_attributes'] = meta[attrs_key]

            return JSONResponse(info)

        return router
